// BasicField.java - LHM - v1.0 - 1997-09-07
// Copyright (c) 1997 Lewis Morton
//
// Class defining a basic holding area, which can be extended for dynamic effects.
// See CollisionField as example displaying elastic collisions.
//
//

package com.lmconsult.elastic;

import java.util.Vector;
import java.awt.*;

/**
 * This is the base class for puck-containing fields.  This provides only framework
 * support for dynamic behavior, which can be provided by extending this class.
 *
 * @author Copyright (c) 1997 Lewis Morton
 * @version 1.0 1997-09-07
 * @see CollisionField
 */

public class BasicField extends FlyoverPanel implements Runnable
{ 
  /**
   * holds the current set of objects known to this field.  Conceptually, this hold dynamic
   * objects, static objects and the walls, in that order.  Handled synchronously.
   */
  Vector things;

  /**
   * number of pucks (eg dynamic objects) in the <code>things</code> vector (stored first)
   */
  int numPucks = 0;
    
  /**
   * number of obstructions (eg static objects) in the <code>things</code> vector [not yet implemented]
   */
  int numObstructions = 0;

  /**
   * number of walls (stored after the pucks and objects in the <code>things</code> vector)
   */
  int numWalls = 0;

  /**
   * system will pause dynamics at each event
   */
  boolean stopMotion = false;
    
  /**
   * freeze dynamics for this many ticks
   */
  int holdTicks;

  /**
   * if time is running in the field's universe
   */
  boolean timeOn = false;

  /**
   * this field is displaying highlighting
   */
  boolean fieldShow = false;

  /**
   * this field is a clonable field
   */
  boolean clonable = false;

  /**
   * this field is a spikable field
   */
  boolean spikable = false;

  /**
   * this field is a resizable field
   */
  boolean canResize = false;
  Dimension hugeSize;
  Dimension tinySize = new Dimension(50,50);
  Dimension trueSize;
  int guard=2;

  /**
   * the Thread for this instance
   */
  Thread myThread = null;

  /**
   * a spare counter
   */
  protected int i;

  /**
   * for double buffering
   */
  protected Graphics g;
  int gx=0;
  int gy=0;

  /**
   * for double buffering
   */
  protected Image im;

  /**
   * size of field- for now only rectangles
   */
  protected Dimension mySize;

  /**
   * the 'real' time
   */
  protected long systemTime;

  /**
   * the system time at the time of the last update tick
   */
  protected long systemLast;

  /**
   * the system time at the time the clock stopped
   */
  protected long timeStop;

  /**
   * the field time at the last update tick
   */
  protected float fieldLast = 0;

  /**
   * the field time we wish to get to in this tick
   */
  protected float tickTime;

  /**
   * the field time right now
   */
  protected float fieldNow;
  
  /**
   *  We're in drag
   */
  boolean dragging=false;

  /**
   * Establish a field, giving the size.  This release constructs a rectangular
   * field filling this area.  A later release may allow arbitrary shapes.
   * @param size the new field's size
   */
  BasicField(Dimension size)
  { mySize = size;
    setSize(mySize);
    setLayout(null);
	things = new Vector(10,10);
    buildWalls();
  }
    
  public void update(Graphics xg)
  { paint(xg);
  }

  /**
   * The field must coordinate the painting services of the enclosed objects.
   */
  public void paint(Graphics xg)
  { Dimension r = getSize();
    if ((r.width>gx) || (r.height>gy))
    { gx=r.width;
	  gy=r.height;
	  im = createImage(gx,gy);
      g = im.getGraphics();
    }
 //   g.setColor(Color.red);
    g.setColor(getForeground());
    Dimension d = getSize();
    g.drawRect(0,0,d.width-1,d.height-1);
 //   g.setColor(Color.yellow);
    g.setColor(getBackground());
    g.fillRect(1,1,d.width-2,d.height-2);
    for (i=0;i<numPucks;i++)
    { Puck puck = (Puck) things.elementAt(i);
      puck.paint(g);
    }
	super.paint(g);
    xg.drawImage(im,0,0,this);
  }

  /**
   * Add a puck to a field.
   * @param newPuck The puck to add.  All current highlighting is set for it.
   */
  synchronized void addPuck(Puck newPuck)
  { things.insertElementAt(newPuck,numPucks);
    numPucks++;
    newPuck.showHighlight(fieldShow,fieldShow);
    newPuck.setOnField(this);
  }

  /**
   * Extract a puck from a field, given a point in the field.
   * @param p A point in the field's address space.
   * @return Puck currently at that point (or null if none)
   */
  synchronized Puck extractPuck(Point p)
  { Puck returnPuck = null;
    for (int i=numPucks-1;i>=0;i--)
    { Puck puck = (Puck) things.elementAt(i);
      Point center = puck.getCenter();
      int d2 = (p.x - center.x)*(p.x - center.x) + (p.y - center.y)*(p.y - center.y);
      int rad = puck.getRadius();
      if ((rad*rad) <= d2) continue;
      returnPuck = puck;
      things.removeElementAt(i);
      numPucks--;
      puck.showHighlight(false,false);
      puck.setOffField();  
      removeRefs(puck);
      break;
    }
    return returnPuck;
  }

  synchronized void removePuck(Puck deprecatedPuck)
  { if (numPucks>0);
    { for (int i=0;i<numPucks;i++)
      { Puck p = (Puck) things.elementAt(i);
        if (p == deprecatedPuck);
        { things.removeElementAt(i);
          numPucks--;
          deprecatedPuck.showHighlight(false,false);
          deprecatedPuck.setOffField();
          removeRefs(deprecatedPuck);
          break;
        }
      }
    }
  }

  public synchronized Puck removeLastPuck()
  { if (numPucks>0)
    { numPucks--;
      Puck temp = (Puck) things.elementAt(numPucks);
      things.removeElementAt(numPucks);
      return temp;
    }
    else return null;
  }
 
  /**
   * Set/unset the basic modes for the field: clonable (picking up a puck
   * causes a duplicate to be left in its place), and spikable (dropping a
   * puck replaces the most recent puck in the field.  We also have the
   * 'canResize', meaning that the field can expand to a given maximum size.
   * The minimum size is the size needed to contain all of the objects in the field. 
   */

   public synchronized void setClonable(boolean b)
   { clonable=b;
   }

   public synchronized boolean getClonable()
   { return clonable;
   }

   public synchronized void setSpikable(boolean b)
   { spikable=b;
   }

   public synchronized boolean getSpikable()
   { return spikable;
   }

   public synchronized void setCanResize(Dimension d)
   { canResize=true;
	 hugeSize=d;
   }
   
   public Dimension getPreferredSize()
   { return mySize;
   }

   public synchronized Dimension getCanResize()
   { return hugeSize;
   }

  /**
   * Establish higlighing modes for pucks in this field.
   * @param b Sets/resets all highlights for every puck in the field.
   * @see Puck
   */
  public synchronized void showDirections(boolean b)
  { fieldShow = b;
    for (int i=0;i<numPucks;i++)
    { Puck p = (Puck) things.elementAt(i);
      p.showHighlight(b,b);
    }
  }

  /**
   * This simple version just builds a wall for each boundary of the field.
   */
  void buildWalls()
  { Rectangle r = getBounds();
    Wall w = new Wall();
    w.horiz = true;
    w.vert = false;
    w.x1 = r.x; w.y1 = r.y; w.x2 = r.x+r.width; w.y2 = r.y;
    things.addElement(w);
    w = new Wall();
    w.horiz = true;
    w.vert = false;
    w.x1 = r.x; w.y1 = r.y+r.height; w.x2 = r.x+r.width; w.y2 = r.y+r.height;
    things.addElement(w);
    w = new Wall();
    w.horiz = false;
    w.vert = true;
    w.x1 = r.x; w.y1 = r.y+r.height; w.x2 = r.x; w.y2 = r.y;
    things.addElement(w);
    w = new Wall();
    w.horiz = false;
    w.vert = true;
    w.x1 = r.x+r.width; w.y1 = r.y; w.x2 = r.x+r.width; w.y2 = r.y+r.height;
    things.addElement(w);
    numWalls = 4;
  }

  /**
   * Check that the given dimensions are wholly contained in this field.
   * @param c the center-point of a object
   * @param r the radius of a (symetrical) object
   */
  public boolean couldEnclose(Point c,int r)
  { Dimension d = getSize();
    if (((c.x+r)>(d.width))
     || ((c.x-r)<(0))
     || ((c.y-r)<(0))
     || ((c.y+r)>(d.height))) return false;
    else return true;
  }

  // Thread mechanics and runnable interface

  public void start()
  { if (myThread == null) 
    { myThread = new Thread(this);
      myThread.start();
    }
  }

  public void stop()
  { myThread = null;
  }

  public void run()
  { while (myThread != null)
    { try {Thread.sleep(100);} catch (InterruptedException e) {}
      if (holdTicks>0)
      { holdTicks--;
        if (holdTicks==0) startTime();
      }
      else if (timeOn) updateDynamics();
      repaint();
    }
    myThread = null;
  }

  // (end Runnable interface)

  /**
   * This plug is called during <code>run()</code>.  Override to provide the dynamic
   * behavior of the field.  This routine must provide any necessary synchronization.
   */
  protected void updateDynamics()
  {
  }

  synchronized void stopTime()
  { if (timeOn)
    { timeStop = System.currentTimeMillis();
      timeOn = false;
    }
    holdTicks = 20;
  }

  synchronized void suspendTime()
  { if (timeOn)
    { timeStop = System.currentTimeMillis();
      timeOn = false;
    }
    holdTicks = 0;
  }

  synchronized void startTime()
  { if (!timeOn)
    { timeStop = 0;
      systemLast = System.currentTimeMillis();
      timeOn = true;
    }
    holdTicks = 0;
  }

  /**
   * Plug for removal of dynamic references to an object
   */
  synchronized void removeRefs(Collidable o)
  {
  }
  
  /**
   * Invoked when we start a resizing operation (presumably someone has clicked
   * the mouse).  Dimension d is a proposed new size.  However, we will never get bigger
   * than the max size set in setCanResize.  We will never be smaller than the biggest
   * object in the field- all of which are moved if necessary.  In any case we never
   * get smaller than tinySize. 
   */
  public void resizeTransient(Dimension d)
  { if ((!dragging) && canResize)
	{ suspendTime();
	  dragging=true;
	  things.setSize(numPucks+numObstructions);
	  int mx = 0;
	  trueSize=new Dimension(tinySize);
	  for (int i=0;i<numPucks;i++)
	  { Puck tp = (Puck) things.elementAt(i);
		int tr = tp.getRadius();
		if (2*(tr+guard)>mx) mx=2*(tr+guard);
	  }  
	  for (int i=numPucks;i<numObstructions+numPucks;i++)
	  { 
	  }
	  if (trueSize.width<mx) trueSize.width = mx;
	  if (trueSize.height<mx) trueSize.height = mx;
	} 
	if (canResize)
	{ int newX=Math.min(hugeSize.width,d.width);
	  int newY=Math.min(hugeSize.height,d.height);
	  newX=Math.max(newX,trueSize.width);
	  newY=Math.max(newY,trueSize.height);
	  mySize = new Dimension(newX,newY);
	  for (int i=0;i<numPucks;i++)
	  { Puck tp = (Puck) things.elementAt(i);
		int tr = tp.getRadius();
		Point tc = tp.getCenter();
		Point xc = tc;
		int oy=tc.y+tr+guard-mySize.height;
		int ox=tc.x+tr+guard-mySize.width;
		oy=Math.max(oy,0);
		ox=Math.max(ox,0);
		if (oy+ox>0) tp.setCenter(new Point(tc.x-ox,tc.y-oy));
	  } 
	  for (int i=numPucks;i<numObstructions+numPucks;i++)
	  { 
	  }
	  setSize(mySize);
	  repaint();
	}
  }

  /**
   * Invoked when we end resizing. Dimension d is the proposed final size
   * for the field.  You may also do all resizing in one operation, by just
   * calling resizeFinal without doing any transients.
   */
  public void resizeFinal(Dimension d)
  { if (canResize)	
	{ resizeTransient(d);
      dragging=false;
	  startTime();
	  buildWalls();
	  refreshField();
	  repaint();
	}
  }
  
  /**
   * You should override this to provide any fix-up of the field following
   * the things Vector being moved and changed.
   */
  public void refreshField()
  {
  }

}
