// CollisionField.java - LHM - v1.0 - 1997-09-07
// Copyright (c) 1997 Lewis Morton
//
// Class defining a field for display of elastic collisions.
//
//

package com.lmconsult.elastic;

import java.util.Vector;
import java.awt.*;

/**
 * This is a field displaying the dynamics of elastic collsions between frictionless,
 * gravity-less pucks.
 *
 * @author Copyright (c) 1997 Lewis Morton
 * @version 1.0 1997-09-07
 */

public class CollisionField extends BasicField implements Runnable
{ 

  /**
   * when a collision occurs before tickTime has expired
   */
  private float stepTime;

  /**
   * flag says more movement still needed on this tick
   */
  private boolean moreToDo;

//debug code
  Collidable c_a;
  Collidable c_b;
//

  /**
   * Establish a field, giving the size.  This release constructs a rectangular
   * field filling this area.  A later release may allow arbitrary shapes.
   * @param size the new field's size
   */
  CollisionField(Dimension size)
  { super(size);
    collisions = new Vector(15,15);
    start();
  }

  /**
   * This plug is called during <code>run()</code>.  Override to provide the dynamic
   * behavior of the field.  This routine must provide any necessary synchronization.
   */
  protected void updateDynamics()
  { do
    { synchronized (this)
      { tickTime = timeToDo();
        moreToDo = false;
        Collision c = getNext(tickTime);
        stepTime = tickTime;
        if (c!=null)
        { stepTime = c.time;
          if (stopMotion) stopTime();
          moreToDo = !stopMotion;
        }
        stepField(stepTime);
        if (c!=null) doCollision(c);
      }
    } while (moreToDo);
    systemLast = timeOn?systemTime:timeStop;
    fieldLast = fieldNow;
    validateCollider();
  }

  protected float timeToDo()			// in synchronized context!
  { systemTime = System.currentTimeMillis();
    float milli = (float) (systemTime-systemLast);
    if (!timeOn) milli = (float) (timeStop-systemLast);
    return milli+fieldLast;
  }

  /**
   * This routine is the dynamic plug- it moves the objects in the
   * <code>things</code> vector.
   * @param t the time (in milliseconds) to update the objects
   * @see CollisionField
   */
  void stepField(float t)
  { float inc = (t - fieldNow);
    for (i=0;i<numPucks;i++)
    { Puck puck = (Puck) things.elementAt(i);
      puck.forward(inc);
    }
    fieldNow = t;
  }

// This is the collider section (this could be an internal class).  We could also make this
// have its own task, but beware of potential synchronizing problems.

  Collision nextCollision;
  boolean unstable;
  Vector collisions;
  long hitWall = 0;
  long hitPuck = 0;
  Solver solver = new Solver();

  synchronized void removeRefs(Collidable o)
  { for (int i=collisions.size()-1;i>=0;i--)
    { Collision c = (Collision) collisions.elementAt(i);
      if ((c.obj1 == o) || (c.obj2 == o))
      { collisions.removeElementAt(i);
      }
    }
    o.invalidateCollisions();
  }

  synchronized Collision getNext(float l)
  { validateCollider();
    Collision nextIs = null;
    if (!collisions.isEmpty())
    { Collision c = (Collision) collisions.firstElement();
      if (c.time <= l)
      { collisions.removeElementAt(0);
        nextIs = c;
      }       
    }
    return nextIs;
  }

  synchronized void validateCollider()
  { Collision n = null;
    for (int i=0;i<numPucks;i++)
    { Puck puck = (Puck) things.elementAt(i);
      if (!puck.collisionsValid)
      { for (int j=0;j<things.size();j++)
        { if (i==j) continue;
          Collidable obj2 = (Collidable) things.elementAt(j);
          float time = setTime(fieldNow, puck, obj2);
          if (time != Float.POSITIVE_INFINITY)
          { Collision t = new Collision();
            t.obj1 = puck;
            t.obj2 = obj2;
            t.time = time;
            insertCollision(t);
          }
        }
      }
      puck.validateCollisions();
    }
    if (collisions.size()>0) n = (Collision) collisions.elementAt(0);
    if (n != null)
    { n.obj1.setHighlight();
      n.obj2.setHighlight();
    }
  }

  void insertCollision(Collision c)
  { int i;
    for (i=0;i<collisions.size();i++)
    { Collision x = (Collision) collisions.elementAt(i);
      if (c.time > x.time) continue;
      break;
    }
    collisions.insertElementAt(c,i);
  }

  float setTime(float timeNow, Collidable obj1, Collidable obj2)
  { Wall wall = null;
    Puck puck = null;
    Puck puckq = null;
    double test = Double.POSITIVE_INFINITY;
    float time = Float.POSITIVE_INFINITY;
    boolean wallType = false;
    if (obj1.getType() == Collidable.WALL)
    { wall = (Wall) obj1;
      puck = (Puck) obj2;
      wallType = true;
    }
    else if (obj2.getType() == Collidable.WALL)
    { wall = (Wall) obj2;
      puck = (Puck) obj1;
      wallType = true;
    }
    if (wallType)
    { if (wall.horiz && (puck.y<wall.y1) && (puck.dy>0)) test = 1000*(wall.y1-puck.y-puck.radius)/puck.dy;
      if (wall.horiz && (puck.y>wall.y1) && (puck.dy<0)) test = 1000*(puck.y-wall.y1-puck.radius)/-puck.dy;
      if (wall.vert && (puck.x<wall.x1) && (puck.dx>0)) test = 1000*(wall.x1-puck.x-puck.radius)/puck.dx;
      if (wall.vert && (puck.x>wall.x1) && (puck.dx<0)) test = 1000*(puck.x-wall.x1-puck.radius)/-puck.dx;
    }
    else
    { puck = (Puck) obj1;
      puckq = (Puck) obj2;
      float d2 = puck.radius + puckq.radius;
      float dext2 = d2 + .5f;
      d2 *= d2;
      dext2 *= dext2;
      float dx = puck.x - puckq.x;
      float dy = puck.y - puckq.y;
      float dist2 = dx*dx + dy*dy;
      float d2x = puck.dx - puckq.dx;
      float d2y = puck.dy - puckq.dy;
//      if ((d2 > dist2) && !(((puck==c_a) & (puckq==c_b)) | ((puckq==c_a) & (puck==c_b)))) 
//      { // stopTime();holdTicks=0;
//        System.out.print(fieldNow);
//        System.out.print(" Underlap! " );
//        System.out.println(Math.sqrt(d2)-Math.sqrt(dist2));
//      }
      float dnorm2 = dext2;
      if (d2 > dist2) test = Double.POSITIVE_INFINITY;
      else
      { if (dext2 > dist2) dnorm2 = d2;
        test=1000 * solver.quad (
          d2x*d2x + d2y*d2y,
          2.0*(dx*d2x + dy*d2y),
          (dist2 - dnorm2));
      }
    }

    if (test == Double.POSITIVE_INFINITY) time = Float.POSITIVE_INFINITY;
    else time = ((float) test) + timeNow;
    return time;
  }

  void doCollision(Collision coll)
//  { synchronized (coll.obj1)
//    { synchronized (coll.obj2)
      { if (coll.obj2.getType() == Collidable.WALL)
        { Puck puck1 = (Puck) coll.obj1;
          Wall wall = (Wall) coll.obj2;
          if (wall.horiz) puck1.dy = -puck1.dy;
          else puck1.dx = -puck1.dx;
          removeRefs(puck1);
          hitWall++;
        }
        else
        { Puck puck1 = (Puck) coll.obj1;
          Puck puck2 = (Puck) coll.obj2;
//
          c_a = puck1;
          c_b = puck2;
//
          float x = puck1.x-puck2.x;
          float y = puck1.y-puck2.y;
          float d2 = (x*x)+(y*y);
          float dx = (puck1.mass*puck1.dx+puck2.mass*puck2.dx)/(puck1.mass + puck2.mass);
          float dy = (puck1.mass*puck1.dy+puck2.mass*puck2.dy)/(puck1.mass + puck2.mass);
          float f1 = 2.0f*(x*(puck1.dx-dx)+y*(puck1.dy-dy))/d2;
          float f2 = 2.0f*(x*(puck2.dx-dx)+y*(puck2.dy-dy))/d2;
          puck1.dx -= f1*x;
          puck1.dy -= f1*y;
          puck2.dx -= f2*x;
          puck2.dy -= f2*y;
//          { float chkd2 = puck1.radius + puck2.radius;
//            chkd2 *= chkd2;
//            System.out.print("Actual distance: " );System.out.println(Math.sqrt(chkd2)-Math.sqrt(d2));
//          }
          removeRefs(puck1);
          removeRefs(puck2);
          hitPuck++;
        }
       coll.obj1.resetHighlight();
       coll.obj2.resetHighlight();
       }
//     }
//   }
  
  	/**
	 * You should override this to provide any fix-up of the field following
	 * the things Vector being moved and changed.
	 * In the dynamic environment, remove all collisions.
	 */
	public void refreshField()
	{ collisions.removeAllElements();
	  for (int i=0;i<things.size();i++)
	  { Collidable c=(Collidable) things.elementAt(i);
		c.collisionsValid=false;
	  }	
	}

 

}
